package DistroMap;
use strict;
use warnings;

use Source::Shared::CLI;
our @ISA=qw/Source::Shared::CLI/;

## TODO: support of eterbuild:
## make a script-converter that converts to map and drop tables
## and update distromap-alt2alt and create distromap from-etherbuild
our $VERSION = 0.395;

my %default_branch = qw/
altlinux	sisyphus
cpan		default
fedora		rawhide
mageia		default
generic		default
jpackage	default
pld		default
pypi		default
rosa		default
/;

# used in Generic2ALT; TODO: make it extendable;
# disabled in ALTLinuxBackport; DirtyHack :(
our $enable_generic_stem_mapper=0;

# TODO: change defaults to generic:generic
# and assure that quick initialization is done if generic
our $default_originrepo='generic';
our $default_originbranch=undef;
our $default_destrepo='generic';
our $default_destbranch=undef;

our ($opt_origin_check_distrodb, $opt_dest_check_distrodb);
#our ($opt_origin_distrodb_version, $opt_dest_distrodb_version);

our @LONGOPT=(
    'originrepo=s' => \$default_originrepo,
    'originbranch=s' => \$default_originbranch,
    'origin-check-distrodb=f' => \$opt_origin_check_distrodb,

    'destrepo=s' => \$default_destrepo,
    'destbranch=s' => \$default_destbranch,
    'dest-check-distrodb=f' => \$opt_dest_check_distrodb,

    # let it be unsupported undocumented option for now;
    'distromap-generic-stem-mapper!' => \$enable_generic_stem_mapper,
    );
our $verbose=0;

my $SMAP;
my $SMAPBACK;
my $BMAP;
my $EXACT_GMAP;
my $APPROX_GMAP;
my $VERSION_UNCOMPARABLE;
my $FLAGS;
# tmp storage for imz@ db extensions
my $EXTRA_UNIMAP={};
#my $eterbuild='/usr/share/eterbuild/pkgrepl';

# public static
sub print_longopt {
    print "DistroMap specific options:\n";
    print "	--originrepo <origin repository name>.".($default_originrepo ? " Default is ".$default_originrepo : ''),"\n";
    print "	--originbranch <origin repository branch name>.",($default_originbranch ? " Default is ".$default_originbranch : ''),"\n";
    print "	--destrepo <destination repository name>.",($default_destrepo ? " Default is ".$default_destrepo : ''),"\n";
    print "	--destbranch <destination repository branch name>.",($default_destbranch ? " Default is ".$default_destbranch : ''),"\n";
    print "	--{origin,dest}-check-distrodb <days>. Warn if DistroDB is older than <days> days. 
	Default is ".
($opt_origin_check_distrodb ? "$opt_origin_check_distrodb days" : 'do not check')." for origin and ".
($opt_dest_check_distrodb ? "$opt_dest_check_distrodb days" : 'do not check')." for dest.\n";

}

sub new {
    my $proto=shift;
    my $class=ref($proto) || $proto;
    my $self = {
	'-originrepo' => $default_originrepo,
	'-originbranch' => $default_originbranch,
	'-destrepo' => $default_destrepo,
	'-destbranch' => $default_destbranch,
	-origin_check_distrodb => $opt_origin_check_distrodb,
	-dest_check_distrodb => $opt_dest_check_distrodb,
#	-origin_distrodb_version => $opt_origin_distrodb_version,
#	-dest_distrodb_version => $opt_dest_distrodb_version,
	@_,
    };
    bless($self,$class);

    # force lowercase of repo and branch names
    foreach (
	'-originrepo',
	'-originbranch',
	'-destrepo',
	'-destbranch',
	) {
	my $val=$self->{$_};
	$self->{$_}=lc($val) if $val;
    }

    $self->{'-originbranch'}=&DistroMap::Repository::select_branch($self->{'-originrepo'},$self->{'-originbranch'});
    $self->{'-destbranch'}=&DistroMap::Repository::select_branch($self->{'-destrepo'},$self->{'-destbranch'});
    my @origin_args=(
	-repo   => $self->{-originrepo},
	-branch => $self->{-originbranch},
	-check_distrodb => $self->{-origin_check_distrodb},
	);
    my @dest_args=(
	-repo   => $self->{-destrepo},
	-branch => $self->{-destbranch},
	-check_distrodb => $self->{-dest_check_distrodb},
	);
    push @origin_args, -distrodb_version => $self->{-origin_distrodb_version} if defined $self->{-origin_distrodb_version};
    push @dest_args, -distrodb_version => $self->{-dest_distrodb_version} if defined $self->{-dest_distrodb_version};
    $self->{ORIGIN}=DistroMap::Repository->new(@origin_args);
    $self->{DEST}=DistroMap::Repository->new(@dest_args);

    unless ($SMAP) {
	$self->__load();
    }
    return $self;
}

sub origin {
    my ($self)=@_;
    return $self->{ORIGIN};
}

sub dest {
    my ($self)=@_;
    return $self->{DEST};
}

# used in distrodb-utils
sub raw_source_map {
    my ($self,$name)=@_;
    return $SMAP->{$name};
}

# to be used in distrodb-utils
sub raw_binary_map {
    my ($self,$name)=@_;
    return $BMAP->{$name};
}

sub multimap_source_dest2origin {
    my ($self,$name)=@_;
    my $map=$SMAPBACK->{$name};
    return @$map if $map;
    # ----------------------------------------------------------------------------------
    # ALT Linux specific built-in filter;
    # should be made an external plug-in against the specified plug-in interface
    if ($enable_generic_stem_mapper) {
	if ($name=~s/^python[23]?-module-//) {
	    return 'python-'.$name;
	}
    }
    # ----------------------------------------------------------------------------------
    return;
}

sub multimap_source_origin2dest {
    my ($self,$name)=@_;
    my $map=$SMAP->{$name};
    return @$map if $map;
    # ----------------------------------------------------------------------------------
    # ALT Linux specific built-in filter;
    # should be made an external plug-in against the specified plug-in interface
    if ($enable_generic_stem_mapper) {
	if ($name!~/^python[23]?-module-/ and $name=~s/^python-//) {
	    return 'python3-module-'.$name;
	}
    }
    # ----------------------------------------------------------------------------------
    return;
}

sub unimap_source_dest2origin {
    my ($self,$name)=@_;
    my $map=$SMAPBACK->{$name};
    return $map->[0] if $map;
    return;
}

sub unimap_source_origin2dest {
    my ($self,$name)=@_;
    my $map=$SMAP->{$name};
    return $map->[0] if $map;
    # ----------------------------------------------------------------------------------
    # ALT Linux specific built-in filter;
    # should be made an external plug-in against the specified plug-in interface
    if ($enable_generic_stem_mapper) {
	if ($name!~/^python[23]?-module-/ and $name=~s/^python-//) {
	    return 'python3-module-'.$name;
	}
    }
    # ----------------------------------------------------------------------------------
    return;
}

sub __exist_in_db {
    return $_[0]->{$_[1]};
}

sub __get_in_db {
    return $_[0]->{$_[1]};
}

sub multimap_binary_origin2dest {
    my ($self,$name)=@_;
    my $map=$BMAP->{$name};
    #print STDERR "multimap_binary_origin2dest: $name - bmap -> ".join(',',@$map)."\n" if $map;
    return @$map if $map;

    my $_dest_provides=$self->dest->_get_db('provides');
    #print STDERR "multimap_binary_origin2dest: $name - provides -> $name\n" if &__exist_in_db($_dest_provides,$name);
    my $whoprovides=&__get_in_db($_dest_provides,$name);
    return $name if $whoprovides and $whoprovides->[0] ne $name;

    my $smap=$self->__stem_map($name);
    #print STDERR "multimap_binary_origin2dest: $name - stem_map_provides -> ".join(',',@$map)."\n" if $smap and &__exist_in_db($_dest_provides,$smap);
    return $smap if $smap and &__exist_in_db($_dest_provides,$smap);
    return $name if $whoprovides;

    # ----------------------------------------------------------------------------------
    # ALT Linux specific built-in filter;
    # should be made an external plug-in against the specified plug-in interface
    if ($enable_generic_stem_mapper) {
	if ($name=~s/-dev(?:el)?$//) {
	    $smap=$self->__stem_map($name);
	    if ($smap) {
		return $smap.'-devel' if &__exist_in_db($_dest_provides,$smap.'-devel');
		return 'lib'.$smap.'-devel' if &__exist_in_db($_dest_provides,'lib'.$smap.'-devel');
	    }
	    return 'lib'.$name.'-devel' if &__exist_in_db($_dest_provides,'lib'.$name.'-devel');
	    if ($name=~s/-libs$// && &__exist_in_db($_dest_provides,'lib'.$name.'-devel')) {
		return 'lib'.$name.'-devel';
	    }
	} elsif ($name=~s/-docs?$//) {
	    $smap=$self->__stem_map($name);
	    for my $prefix ('','lib') {
		for my $suffix (qw/-doc -docs/) {
		    return $prefix.$smap.$suffix if $smap and &__exist_in_db($_dest_provides,$prefix.$smap.$suffix);
		}
	    }
	} elsif ($name=~s/-firmware$//) {
	    $smap=$self->__stem_map($name);
	    return 'firmware-'.$smap if $smap;
	    return 'firmware-'.$name;
	} elsif ($name=~s/-python$// and $name ne 'rpm-build-python' and $name ne 'texlive-python') {
	    $smap=$self->__stem_map($name);
	    return 'python3-module-'.$smap if $smap;
	    return 'python3-module-'.$name;
	} elsif ($name=~s/-python3$// and $name ne 'rpm-build-python3' and $name ne 'texlive-python3') {
	    $smap=$self->__stem_map($name);
	    return 'python3-module-'.$smap if $smap;
	    return 'python3-module-'.$name;
	}
    }
    # ----------------------------------------------------------------------------------
    return;
}

# tmp api for imz@ db extensions (this util only)
sub db_unimap_extra_origin2dest {
    my ($self,$dbname,$name)=@_;
    my $db=$EXTRA_UNIMAP->{$dbname};
    unless ($db) {
	$db={};
	$EXTRA_UNIMAP->{$dbname}=$db;
	my $originrepo=$self->{'-originrepo'};
	my $destrepo=$self->{'-destrepo'};
	my $originbranch=$self->{'-originbranch'};
	my $destbranch=$self->{'-destbranch'};
	foreach my $file ($self->__find_2map_files($originrepo,$originbranch,$destrepo,$destbranch,$dbname)) {
	    &__load_unimap_as($file,$db);
	}
    }
    return $db->{$name};
}


# by imz@ request (this util only)
sub db_multimap_binary_origin2dest {
    my ($self,$name)=@_;
    my $map=$BMAP->{$name};
    #print STDERR "multimap_binary_origin2dest: $name - bmap -> ".join(',',@$map)."\n" if $map;
    return @$map if $map;
    return;
}

sub unimap_exact_group_origin2dest {
    my ($self,$name)=@_;
    return $EXACT_GMAP->{$name};
}

sub unimap_approximate_group_origin2dest {
    my ($self,$name)=@_;
    return $APPROX_GMAP->{$name};
}

sub versions_are_uncomparable {
    my ($self,$name)=@_;
    return $VERSION_UNCOMPARABLE->{$name};
}

sub flag {
    my ($self,$flag,$name)=@_;
    return $FLAGS->{$flag}->{$name};
}

########## PROTECTED #################

### hack ...
sub die_if_no_distromap {
    my ($self, $type)=@_;
    $type//='source';
    my $dir=&_get_distromap_dir($self->{-originrepo},$self->{-originbranch},$self->{-destrepo},$self->{-destbranch},$type);
    unless ($dir) {
	my $originrepo=$self->{-originrepo};
	my $originbranch=$self->{-originbranch};
	my $destrepo=$self->{-destrepo};
	my $destbranch=$self->{-destbranch};
	my $gitdir=&__local_distromap_git_dir($originrepo,$originbranch,$destrepo,$destbranch,$type);
	my $cachedir=$ENV{'HOME'}."/.cache/distromap/$originrepo/$originbranch/$destrepo/$destbranch/$type";
	my $systemdir="/usr/share/distromap/$originrepo/$originbranch/$destrepo/$destbranch/$type";
	print STDERR "Looking for:\n";
	print STDERR " gitdir=$gitdir\n";
	print STDERR " cachedir=$cachedir\n";
	print STDERR " systemdir=$systemdir\n";
    }
    die "Fatal: distromap-$self->{-originrepo}-$self->{-originbranch}-$self->{-destrepo}-$self->{-destbranch} is required." unless $dir;
}

########## END PROTECTED #################


package DistroMap::Repository;

use strict;
use warnings;
use Source::Shared::CLI;
our @ISA=qw/Source::Shared::CLI/;

our $verbose;

our ($opt_check_distrodb,$opt_distrodb_version);

our @LONGOPT=(
    'check-distrodb=f' => \$opt_check_distrodb,
    );

my $REPODATA={};

# public static
sub print_longopt {
    print "DistroDB specific options:\n";
    print "	--check-distrodb <days>. Warn if DistroDB is older than <days> days. Default is ".($opt_check_distrodb ? "$opt_check_distrodb days" : 'do not check').".\n";
}

sub new {
    my $proto=shift;
    my $class=ref($proto) || $proto;
    my $self = {
	-repo => 'generic',
	-branch => undef,
	-check_distrodb => $opt_check_distrodb,
	-distrodb_version => $opt_distrodb_version,
	@_,
    };
    bless($self,$class);

    # force lowercase of repo and branch names
    foreach (
	'-repo',
	'-branch',
	) {
	my $val=$self->{$_};
	$self->{$_}=lc($val) if $val;
    }

    $self->{'-branch'}=&select_branch($self->{'-repo'},$self->{'-branch'});

    if ($verbose and $verbose > 2) {
	print STDERR "DistroMap::Repository: new repo=",$self->{'-repo'},", branch=",$self->{'-branch'},"\n";
    }

    if ($self->{-check_distrodb}) {
	$self->__check_distrodb();
    }
    my $repo=$self->{'-repo'};
    my $branch=$self->{'-branch'};
    $self->{'DATA'}=&__get_repodata($repo,$branch);

    return $self;
}

sub __check_distrodb {
    my ($self)=@_;
    my $repo=$self->{'-repo'};
    my $branch=$self->{'-branch'};
    # Live Fast Die Young - warn if repo is older than -check_distrodb days
    my $dir=$ENV{'HOME'}."/.cache/distrodb/$repo/$branch";
    my $file_vmin=$dir.'/.compat';
    my $file_vmax=$dir.'/.version';
    if (not -e $dir or not -e $file_vmax) {
	die "FATAL: dynamic DistroMap repository data for $repo:$branch is missing!\n",
	    $repo eq 'altlinux' ? "run distrodb-update-repocop-db-$repo-$branch!\n":
	    "run distrodb-helper-mkdistrodb\n";
    } elsif (-M $file_vmax > $self->{-check_distrodb}) {
	    warn "WARNING: dynamic repository data for $repo:$branch is too old.\n";
	if ($repo eq 'altlinux') {
	    warn "WARNING: run distrodb-update-repocop-db-$repo-$branch\n";
	} else {
	    warn "WARNING: run distrodb-helper-mkdistrodb\n";
	}
    }

    # check distrodb versioning
    my $version_req=$self->{-distrodb_version};
    # 0==skip version checking
    if ($version_req) {
	die "FATAL: damaged DistroDB repository data: not found: $file_vmin or $file_vmax\n" if not -e $file_vmin or not -e $file_vmax;
	my $version_min=&__get_first_line_of_file($file_vmin);
	my $version_max=&__get_first_line_of_file($file_vmax);
	die "FATAL: Version $version_req is required, but DistroDB repository data for $repo:$branch is not versioned!\n" if not defined $version_max or not defined $version_min;
	die "FATAL: Version $version_req is required, but DistroDB repository data for $repo:$branch has version $version_max (compatible with version >= $version_min)!\nTry to run distrodb-update-repocop-db-$repo-$branch!\n" if $version_max < $version_req or $version_min > $version_req;
    }
}

sub select_branch {
    my ($repo, $branch)=@_;
    return $branch if defined $branch;
    $branch = $default_branch{$repo};
    return $branch if defined $branch;
    return 'default';
}

# also used in package RPM::Source::Convert::Plugins::DistroMap::Packages;
sub translate {
    my ($self,$db,$name)=@_;
    my $map=$self->_get_db($db)->{$name};
    return unless $map;
    return wantarray ? @$map : $map->[0];
}

sub exist {
    my ($self,$db,$name)=@_;
    return $self->_get_db($db)->{$name};
}

# TODO
#package DistroMap::Repository::LazyLoader;


########## PRIVATE #################
package DistroMap::Repository;

use strict;
use warnings;

sub __get_repodata {
    my @path=@_;
    my $node=$REPODATA;
    while (@path>0) {
	my $step=shift @path;
	$node->{$step}={} unless defined $node->{$step};
	$node=$node->{$step};
    }
    return $node;
}

sub _get_db {
    my ($self,$type)=@_;
    my $MAP=$self->{'DATA'}->{$type};
    return $MAP if $MAP;
    $MAP={};
    $self->{'DATA'}->{$type}=$MAP;
    $self->__load_type($type,$MAP);
    return $MAP;
}

sub __load_type {
    my ($self,$type,$MAP)=@_;
    my $repo=$self->{'-repo'};
    return if $repo eq 'generic';
    my $branch=$self->{'-branch'};
    print STDERR "DistroMap: reading $repo $branch $type\n" if $verbose and $verbose > 1;
    if (length($type)>5 and substr($type,0,5) eq 'list.') {
	#print STDERR "__load_type: loading $type as list\n";
	foreach my $file (&__glob_1map_files($repo,$branch,$type)) {
	    &DistroMap::__load_list_as($file,$MAP);
	}
    } else {
	foreach my $file (&__glob_1map_files($repo,$branch,$type)) {
	    &DistroMap::__load_map_as($file,$MAP);
	}
    }
}

sub __glob_1map_files {
    my ($repo,$branch,$type)=@_;
    my @out;
    my $homedir=$ENV{'HOME'};
    my $cachedir="$homedir/.cache/distrodb/$repo/$branch/$type";
    my $systemdir="/usr/share/distrodb/$repo/$branch/$type";
    my $devdir="$homedir/src/repo/distrodb-static-${repo}-${branch}.git/data/$type";
    my $dir=&DistroMap::__select_most_fresh_dir($cachedir, $systemdir, $devdir);
    push @out, glob "$dir/*.{txt,tsv}" if $dir;
    return @out;
}

sub __get_first_line_of_file {
    my ($file)=@_;
    open (my $fn, '<', $file) || die "can't open $file: $!";
    my $line=<$fn>;
    chomp $line;
    close ($fn);
    return $line;
}

########## END PRIVATE #################


########## PRIVATE #################
package DistroMap;

use strict;
use warnings;

sub __stem_map {
    my ($self,$name)=@_;
    my $map=$SMAP->{$name};
    return $map->[0] if $map and $map->[0] ne $name;
    return;
}

sub __load {
    my ($self)=@_;
    $SMAP={};
    $BMAP={};
    $SMAPBACK={};
    $EXACT_GMAP={};
    $APPROX_GMAP={};
    $VERSION_UNCOMPARABLE={};
    $FLAGS={
	'flags/source/noversion' => $VERSION_UNCOMPARABLE,
	'flags/source/exclude' => {},
	'flags/binary/noepoch' => {},
	'flags/binary/noversion' => {},
	'flags/binary/exclude' => {},
    };
    my $originrepo=$self->{'-originrepo'};
    my $destrepo=$self->{'-destrepo'};
    my $originbranch=$self->{'-originbranch'};
    my $destbranch=$self->{'-destbranch'};

    return if $originrepo eq 'generic' and $destrepo eq 'generic';
    return if $originrepo eq $destrepo and $originbranch eq $destbranch;

    foreach my $file ($self->__find_2map_files($originrepo,$originbranch,$destrepo,$destbranch,'source')) {
	&__load_map_as($file,$SMAP,$SMAPBACK);
    }
    foreach my $file ($self->__find_2map_files($destrepo,$destbranch,$originrepo,$originbranch,'source')) {
	&__load_map_as($file,$SMAPBACK);
    }
    foreach my $file ($self->__find_2map_files($originrepo,$originbranch,$destrepo,$destbranch,'binary')) {
	&__load_map_as($file,$BMAP);
    }

    foreach my $file ($self->__find_2map_files($originrepo,$originbranch,$destrepo,$destbranch,'group-strict')) {
	&__load_unimap_as($file,$EXACT_GMAP);
    }
    foreach my $file ($self->__find_2map_files($originrepo,$originbranch,$destrepo,$destbranch,'group-approx')) {
	&__load_unimap_as($file,$APPROX_GMAP);
    }

    foreach my $file ($self->__find_2map_files($originrepo,$originbranch,$destrepo,$destbranch,'noversion')) {
	&__load_list_as($file,$VERSION_UNCOMPARABLE);
    }

    foreach my $flag (qw!flags/source/exclude flags/source/noversion  flags/binary/noepoch flags/binary/noversion flags/binary/exclude!) {
	foreach my $file ($self->__find_2map_files($originrepo,$originbranch,$destrepo,$destbranch,$flag)) {
	    &__load_list_as($file,$FLAGS->{$flag});
	}
    }
}

sub __eabbr {
    my $a=shift;
    return $a;
}

sub __etype {
    my $a=shift;
    return 'pkgrepl' if $a eq 'binary';
    return 'grprepl' if $a eq 'group-approx';
    return;
}

my %inherit_generic=map {$_=>1} qw/fedora suse mageia mandriva pld rosa/;

sub __find_2map_files {
    my ($self, $originrepo,$originbranch,$destrepo,$destbranch,$type)=@_;
    my @out;
    # TODO: implement chains
    if ($originrepo eq 'jpackage') {
	push @out, $self->__glob_2map_files('generic','default',$destrepo,$destbranch,$type);
	push @out, $self->__glob_2map_files('fedora','rawhide',$destrepo,$destbranch,$type);
    } elsif ($inherit_generic{$originrepo}) {
	push @out, $self->__glob_2map_files('generic','default',$destrepo,$destbranch,$type);
    }

    my $origindef=$DistroMap::default_branch{$originrepo};
    my $destdef=$DistroMap::default_branch{$destrepo};
    push @out, $self->__glob_2map_files($originrepo,$origindef,$destrepo,$destdef,$type)
	if $origindef and $origindef ne $originbranch and $destdef and $destdef ne $destbranch;
    push @out, $self->__glob_2map_files($originrepo,$origindef,$destrepo,$destbranch,$type)
	if $origindef and $origindef ne $originbranch and (not $destdef or $destdef ne $destbranch);
    push @out, $self->__glob_2map_files($originrepo,$originbranch,$destrepo,$destdef,$type)
	if $destdef and $destdef ne $destbranch and (not $origindef or $origindef ne $originbranch);
    push @out, $self->__glob_2map_files($originrepo,$originbranch,$destrepo,$destbranch,$type);
    return @out;
}

sub __glob_2map_files {
    my ($self, $originrepo,$originbranch,$destrepo,$destbranch,$type)=@_;
    my @out;
    # if ($originrepo eq 'altlinux' && $originbranch eq 'sisyphus'&& -d $eterbuild) {
    # 	my $drepo=&__eabbr($destrepo);
    # 	my $dbranch=&__eabbr($destbranch);
    # 	my $etype=&__etype($type);
    # 	push @out, grep {-f $_} "$eterbuild/$etype.$drepo", "$eterbuild/$etype.$drepo.$dbranch" if $etype;
    # }
    my $dir=&_get_distromap_dir($originrepo,$originbranch,$destrepo,$destbranch,$type);
    push @out, glob "$dir/*.{txt,tsv}" if $dir;
    return @out;
}

sub _get_distromap_dir {
    my ($originrepo,$originbranch,$destrepo,$destbranch,$type)=@_;
    my $gitdir=&__local_distromap_git_dir($originrepo,$originbranch,$destrepo,$destbranch,$type);
    my $cachedir=$ENV{'HOME'}."/.cache/distromap/$originrepo/$originbranch/$destrepo/$destbranch/$type";
    my $systemdir="/usr/share/distromap/$originrepo/$originbranch/$destrepo/$destbranch/$type";
    return &__select_most_fresh_dir($gitdir, $cachedir, $systemdir);
}

sub __local_distromap_git_dir {
    my ($originrepo,$originbranch,$destrepo,$destbranch,$type)=@_;
    $destrepo='alt' if $destrepo eq 'altlinux';
    return $ENV{HOME}.'/src/repo/distromap-'.$originrepo.'2'.$destrepo.'.git/'.$type;
}

sub __select_most_fresh_dir {
    #print STDERR "DistroMap: most_fresh_dir: args: @_\n" if $verbose and $verbose > 4;
    my @indir=grep {-d $_} @_;
    #print STDERR "DistroMap: most_fresh_dir: exists: @indir\n" if $verbose and $verbose > 4;
    return unless @indir;
    my $dir=shift @indir;
    while (scalar @indir) {
	my $newdir=shift @indir;
	$dir=$newdir if -M $dir > -M $newdir;
    }
    #print STDERR "DistroMap: most_fresh_dir: selected: $dir\n" if $verbose and $verbose > 4;
    return $dir;
}

sub __split_by_space {
    return split (/\s+/, $_[0]);
} 
# sub __split_etersoft {
#     return split (/|/, $_[0]);
# } 
sub __split_tsv {
    return split (/ *\t+ */, $_[0]);
} 

sub __load_map_as {
    my ($file,$MAP,$MAPBACK)=@_;
    print STDERR "DistroMap: loaded map $file\n" if $verbose and $verbose > 2;
    my $splitter=\&__split_by_space;
    $splitter=\&__split_tsv if $file=~/tsv$/;
    # $splitter=\&__split_etersoft if $file=~m!/eterbuild/!;
    open (my $fn, '<', $file) || die "can't open $file: $!";
    local $_;
    while (<$fn>) {
	chomp;
	next if /^#/;
	next if /^\s*$/;
	warn "$file: endspace found" if s/\s+$//;
	my @map=grep {$_} &$splitter($_);
	my $from=shift @map;
	die "$file: invalid map $_" unless scalar @map>0;
	$MAP->{$from}=\@map;
	map {push @{$MAPBACK->{$_}}, $from} @map if $MAPBACK;
    }
    close ($fn);
}

sub __load_unimap_as {
    my ($file,$MAP,$MAPBACK)=@_;
    print STDERR "DistroMap: loaded map $file\n" if $verbose and $verbose > 2;
    my $split=qr'\s+';
    $split=qr' *\t+ *' if $file=~/tsv$/;
    # $split=qr'|' if $file=~m!/eterbuild/!;
    open (my $fn, '<', $file) || die "can't open $file: $!";
    local $_;
    while (<$fn>) {
	chomp;
	next if /^#/;
	next if /^\s*$/;
	my @map=grep {$_} split (/$split/, $_);
	warn "$file: extra args in unimap: @map\n" if @map >2;
	my $from=shift @map;
	die "$file: invalid map $_" unless @map;
	$MAP->{$from}=$map[0];
	$MAPBACK->{$map[0]}=[$from] if $MAPBACK and (scalar @map == 1);
    }
    close ($fn);
}

sub __load_list_as {
    my ($file,$LIST)=@_;
    print STDERR "DistroMap: loaded list $file\n" if $verbose and $verbose > 2;
    open (my $fn, '<', $file) || die "can't open $file: $!";
    local $_;
    while (<$fn>) {
	chomp;
	next if /^#/;
	next if /^\s*$/;
	s/\s*$//;
	$LIST->{$_}=1;
    }
    close ($fn);
}

########## END PRIVATE #################


1;

=head1	NAME

DistroMap - DistroMap database query library.

=head1	DESCRIPTION

DistroMap query library
provides an object interface to DistroMap database.

=head1	AUTHOR

Written by Igor Vlasenko <viy@altlinux.org>.

=head1	COPYING

Copyright (c) 2010-2023 Igor Vlasenko, ALT Linux Team.

This is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

=cut

